@@ -48,6 +48,11 @@ def get_download_url() -> str:
     if sys.platform == "win32":
         return gh_prefix + "pdfstitcher.exe"
     elif sys.platform == "darwin":
-        return gh_prefix + "PDFStitcher-Installer.dmg"
+        import platform
+
+        if platform.machine() == "arm64":
+            return gh_prefix + "PDFStitcher-InstallerARM64.dmg"
+        else:
+            return gh_prefix + "PDFStitcher-InstallerX64.dmg"
     else:
         return utils.FLATHUB_HOME
